/*
 *  Periodically takes transactions from its queue and inserts them in another queue or deletes them.
 *
 */

import cz.zcu.fav.kiv.jsim.*;
import cz.zcu.fav.kiv.jsim.unsecure.*;

public class Server extends JSimProcess {
    
    private double                  mu;
    private double                  p;
    private QueueWithServer         queueIn;
    private QueueWithServer         queueOut;
    private ClientData              data;
    
    private boolean                 distributedSim;
    private String                  receiverName;
    private int                     receiverGate;
    private int                     counter;
    private double                  transTq;
    public DistributedSimulation    simulation;
    
    /**
     *  Use this Constructor for distributed simulation but not for
     *  output point, transaction are throw out or put to other server queue
     */
    public Server(String name,DistributedSimulation  sim, double parMu, double parP, QueueWithServer parQueueIn, QueueWithServer parQueueOut,
    ClientData parData) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException {
        
        super(name, sim);
        mu              = parMu;
        p               = parP;
        queueIn         = parQueueIn;
        queueOut        = parQueueOut;
        data            = parData;
        distributedSim  = false;
        counter         = 0;
        transTq         = 0.0;
        
    } // constructor
    
    /**
     *  Constructor for distributed simulation for point where you can store events to output queue
     *  and send to other simulation node
     */
    public Server(String name,DistributedSimulation  sim, double parMu, double parP, QueueWithServer parQueueIn, QueueWithServer parQueueOut,
    ClientData parData, String parReceiverName, int parGateNumber) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException {
        
        super(name, sim);
        mu              = parMu;
        p               = parP;
        queueIn         = parQueueIn;
        queueOut        = parQueueOut;
        data            = parData;
        distributedSim  = true;
        receiverName    = parReceiverName;
        receiverGate    = parGateNumber;
        counter         = 0;
        transTq         = 0.0;
        
    } // constructor
    
    
    protected void life() {
        Transaction   t;
        JSimLink      link;
        
        try {
            while (true) {
                if (queueIn.empty()) {
     /*
      *  If we have nothing to do, we sleep.
      */
                    //message(getProcessName() + ": The queue is empty, going to sleep.");
                    passivate();
                    //message(getProcessName() + ": Something has happened with my queue, woken up.");
                }
                else {
      /*
       *  Simulating hard work here...
       *
       *
       */
                    hold(JSimSystem.negExp(mu));
                    link = queueIn.first();
      /*
       *  This condition is nessary becuse when undo comin the link is erased
       *
       */
                    if (link != null) {
                        
                        t = (Transaction) link.getData();
                        data.WriteToSimFile(getName() + "   : CT :" + myParent.getCurrentTime()+ ": CTS :" + t.getCreationTime());
      /*
       *  Now we must store all values of this server for future, When undo coming and we need restoring statistick values
       */               
                        data.addSimUndoItem(this, myParent.getCurrentTime(), t.getCreationTime(), transTq, queueIn.getSumTwRemoved(), queueIn.getNoOfItemsRemoved(), queueIn.getSumLwUntilLastChange());
                        
     /*
      *  Now we must decide whether to throw the transaction away or to insert it into another queue or send to distributed simulation
      */
                        if (JSimSystem.uniform(0.0, 1.0) > p) {
                            //message(getName() + ": Throwing the transaction away.");
                            t = (Transaction) link.getData();
                            counter++;
                            transTq += myParent.getCurrentTime() - t.getCreationTime();
                            this.throwOutLink(link);
                        }
                        else {
                            //message(getProcessName() + ": Inserting the transaction into another queue.");
                            if (distributedSim == false) { /* switch is distributed simulation  ?*/
                                if (queueOut != null)  {
                                    link.out();
                                    link.into(queueOut);
                                    if (queueOut.getServer().isIdle()) {
                                        //message(getProcessName() + ": The other queue's server is idle, waking it up.");
                                        queueOut.getServer().activate(myParent.getCurrentTime());
                                    } // if isIdle
                                }
                                else {
     /*
      * Transaction is Throw out of simulation
      */                            //message(getName() + ": Throwing the transaction away.");
                                    t = (Transaction) link.getData();
                                    counter++;
                                    transTq += myParent.getCurrentTime() - t.getCreationTime();
                                    this.throwOutLink(link);
                                }
                            }
                            else {
     /*
      * Transaction is Throw to distributed simulation
      */
                                t = (Transaction) link.getData();
                                
                                // this branch made straggler message for simulation named "vlk"
                                // if you want you can change name of simulation or skip this condition 
                                if (data.simName.toString().equals("vlk") == true){
                                    if (myParent.getCurrentTime()<500.0)
                                        data.putToOutputQueue(myParent.getCurrentTime() + 500.0, t.getCreationTime()+500.0,  this.receiverName, this.receiverGate);
                                }
                                else {
     /*
      * There we store transaction to output queue
      */
                                    data.putToOutputQueue(myParent.getCurrentTime(), t.getCreationTime(), this.receiverName, this.receiverGate);
                                }
                                this.throwOutLink(link);
                            } // else local / distributed simulation
                        } // else throw away / insert
                    }
                } // else queue is empty / not empty
            } // while
        } // try
        catch (JSimException e) {
            e.printStackTrace();
            e.printComment(System.err);
        }
    } // life
    public int getCounter() {
        return counter;
    } // getCounter
    
   /**
    *   This method throw out link out of system        <BR>
    */ 
    private void throwOutLink(JSimLink  parLink){
        try {
            if (parLink != null) {
                parLink.out();
                parLink = null;
            }
        } // try
        catch (JSimException e) {
            e.printStackTrace();
            e.printComment(System.err);
        }
        
        return;
    }
    
    /**
     *  This method is call vhen undo operation is comin            <BR>
     *  and restore all values from time when simulation was        <BR>
     *  in stable state                                             <BR>
     *
     *  please don't call it direct                                 <BR>
     */
    
    public void restore(JSimProcess ParProcess){
        TimeItem t_item;
        
        // there we load last correct statistic data
        t_item = data.getSimItem(ParProcess);
        // there we insert information to simulation
        transTq = t_item.transTq;
        queueIn.setNoOfItemsRemoved((long) t_item.linkNoOfItemsRemoved);
        queueIn.setSumLwUntilLastChange(t_item.linkSumLwUntilLastChange);
        queueIn.setSumTWRemoved(t_item.linkSumTwRemoved);
        
        message(getName() + ": Restore states from CreatinTS :" + t_item.creationTime + " : "+ t_item.linkNoOfItemsRemoved +" : " + t_item.linkSumLwUntilLastChange);
        
        return;
    }
    
    
    public double getTransTq() {
        return transTq;
    } // getTransTq
} // class Server

