/*
 *	Copyright (c) 2000 - 2004, University of West Bohemia,
 *	                           Faculty of Applied Sciences,
 *	                           Department of Computer Science and Engineering,
 *	                           DSS Group.
 *	                           Pilsen, Czech Republic.
 *
 *	Please read the licence agreement carefully before using any part of this software.
 */
package cz.zcu.fav.kiv.jsim;

import cz.zcu.fav.kiv.jsim.gui.*;

import java.util.*;
import javax.swing.*;

/**
 *	A JSimProcess is an independent unit of code, having its own data. Its code is placed in
 *	the life() method. Processes have possibilities to schedule themselves as well as other processes,
 *	they can also cancel another process's schedule. A process must belong to a simulation.
 *	During a simulation execution, processes are interleaved in a pseudo-parallel manner, i.e.
 *	just one process is running at the same time. When the currently running process invokes
 *	hold(), passivate(), or reactivate(), it is suspended and another process can be executed instead.
 *
 *	@author		<a href="http://home.zcu.cz/~jkacer">Jarda KAER</a>
 *	@author		Pavel DOMECK
 *
 *	@version	J-Sim version 0.3.0
 *
 *	@since		J-Sim version 0.0.1
 */
public class JSimProcess extends Thread implements JSimDisplayable, Comparable
{
	// Process states

	/**
	 *	A process is in STATE_NEW state if it has been created but not started and not scheduled yet.
	 */
	public static final int STATE_NEW = 0;
	/**
	 *	A process is passive if it has been started and it has no event in the calendar.
	 */
	public static final int STATE_PASSIVE = 1;
	/**
	 *	A process is scheduled if it has an event in the calendar so it will run in the future.
	 */
	public static final int STATE_SCHEDULED = 2;
	/**
	 *	A process is active if it is just now running.
	 */
	public static final int STATE_ACTIVE = 3;
	/**
	 *	A process is terminated if it has reached the end of its life() method.
	 */
	public static final int STATE_TERMINATED = 4;

	// Methods passivate(), hold(), and reactivate() will ignore first INT_REQUESTS_TO_IGNORE requests
	// to warn the programmer.
	public static final int INT_REQUESTS_TO_IGNORE  = 1;

//--------------------------------------------------------------------------------------------------

	/**
	 *	The process's number. It is unique for all process within the same simulation.
	 */
	private final long myNumber;

	/**
	 *	This process's state -- new, passive, scheduled, active or terminated.
	 */
	private int myState;

	/**
	 *	The simulation that this process is a part of. Protected because child classes will want to use it.
	 */
	protected final JSimSimulation myParent;

	/**
	 *	Lock used when switching between the simulation and the process.
	 *	Final -- can be just created, not changed afterwards.
	 */
	private final Object myOwnLock;

	/**
	 *	Flag saying whether this process has been started. Even a new thread can be scheduled,
	 *	we must distinguish between being new and being started.
	 */
	private boolean haveIBeenStarted;

	/**
	 *	Flag saying whether this process has been interrupted by its simulation and should terminate.
	 */
	private boolean shouldTerminate;

	/**
	 *	Number of terminating warnings this process has got since it received an interrupt signal.
	 *	It should be at most 1.
	 */
	private int requestsToTerminate;

	/**
	 *	The simulation time that the process is scheduled for (if it is scheduled).
	 */
	private double scheduleTime;

//--------------------------------------------------------------------------------------------------

	/**
	 *	Creates a new process having a name and belonging to a simulation. If the simulation is already
	 *	terminated or there is no free number the process is not created and an exception is thrown out.
	 *	All processes must have their parent simulation object specified. If not, an exception is thrown, too.
	 *	After the creation, the process is not scheduled and therefore will not run unless explicitely activated
	 *	by another process or the main program using activate().
	 *
	 *	@param name	Name of the process being created.
	 *	@param parent	Pparent simulation.
	 *
	 *	@exception JSimSimulationAlreadyTerminatedException	This exception is thrown out if the simulation
	 *								has already been terminated.
	 *	@exception JSimInvalidParametersException		This exception is thrown out if no parent was
	 *								specified.
	 *	@exception JSimTooManyProcessesException		This exception is thrown out if no other process
	 *								can be added to the simulation specified.
	 *	@exception JSimKernelPanicException			This exception is thrown out if the simulation
	 *								is in a unknown state. Do NOT catch this exception!
	 */
	public JSimProcess(String name, JSimSimulation parent) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
	{
		super(name);

		// We must know our simulation context
		if (parent == null)
			throw new JSimInvalidParametersException("JSimProcess.JSimProcess(): parent");
		else
			myParent = parent;

		myNumber = myParent.getFreeProcessNumber();
		if (myNumber == JSimSimulation.NEW_PROCESS_FORBIDDEN)
			throw new JSimTooManyProcessesException("JSimProcess.JSimProcess()");

		// Final -- Must be initialized before branching
		myOwnLock = new Object();

		// Sometimes, it is not allowed to create a new process. Let's see...
		switch (myParent.getSimulationState())
		{
			// OK, let's create a new J-Sim process
			case JSimSimulation.SIMULATION_NOT_STARTED:
			case JSimSimulation.SIMULATION_IN_PROGRESS:
				try
				{
					myState = STATE_NEW;
					myParent.addProcess(this);
					haveIBeenStarted = false;
					shouldTerminate = false;
					requestsToTerminate = 0;
					scheduleTime = JSimCalendar.NEVER;
				} // try
				catch (JSimInvalidParametersException e)
				{
					e.printStackTrace();
					e.printComment(System.err);
					throw e; // Rethrow the exception.
				} // catch
				break;

			// Notify the user by throwing out an exception
			case JSimSimulation.SIMULATION_TERMINATED:
				throw new JSimSimulationAlreadyTerminatedException(name);

			// Something strange must have happened. Ufff...
			default:
				myParent.talk("NEW PROCESS: The simulation is in an unknown state. J-Sim Kernel in panic. Please report.");
				throw new JSimKernelPanicException();
		} // switch
	} // constructor

//--------------------------------------------------------------------------------------------------

	/**
	 *	Stops the process and waits until it is woken up by a signal. When a new process is started, it always
	 *	blocks here and waits until the main thread switches to it when its turn comes.
	 *	You should never use this method. It is called automatically from run().
	 *
	 *	@exception JSimProcessDeath	This internal J-Sim exception is thrown out when the simulation is
	 *					shutting down. You should NEVER catch it!
	 */
	private final void getReady()
	{
		synchronized(myOwnLock)
		{
			try // outer try
			{
				// The process might have been already scheduled with activate().
				// In that case, let it be so. Otherwise, change the state from NEW to PASSIVE.
				if (getState() != STATE_SCHEDULED)
					setProcessState(STATE_PASSIVE);

				while (myParent.getRunningProcess() != this)
				{
					try  // inner try
					{
						myOwnLock.wait();
					} // inner try
					catch (InterruptedException e1)
					{
						// We must know this not to enter the life() method.
						// The simulation is shutting down.
						shouldTerminate = true;
						requestsToTerminate++;

						// We must not fall asleep again. When the simulation deletes all
						// threads, the running process is set to NOBODY.
						// Without break, we would sleep forever because every process is
						// interrupted just once during the shutdown.
						break;
					} // catch
				} // while

				// If we have to die this exception is propagated till run().
				// Otherwise we will normally continue with life().
				if (shouldTerminate)
				{
					myParent.talk(getName() + " GET_READY: Killed, never run.");
					throw new JSimProcessDeath();
				} // if
				else
					setProcessState(STATE_ACTIVE);
			} // outer try
			catch (JSimInvalidProcessStateException e2)
			{
				e2.printStackTrace();
				e2.printComment(System.err);
				myParent.talk(getName() + " GET_READY: A bug appeared. The process is running! Expect simulation malfunction.");
			} // catch
		} // synchronized
	} // getReady

	/**
	 *	Informs the parent simulation about its termination and switches back to the main thread.
	 *	You should never use this method.
	 */
	private final void finish()
	{
		synchronized(myOwnLock)
		{
			try
			{
				// If we are not killed but just have run to the end we must inform our parent
				if (!shouldTerminate)
				{
					myParent.deleteProcess(this);
					myParent.switchToNobody();
					myOwnLock.notify();
					myParent.talk(getName() + " FINISH: Dying and switching to the main thread.");
				} // if
				else
				{
					// If we are killed by our parent everythink is OK, we are not doing anything.
					// The simulation has done everything in shutdown().
					myParent.talk(getName() + " FINISH: Killed by the main thread, no activity.");
				} // else

				setProcessState(STATE_TERMINATED);
				// And now, we can peacefully die...
			} // try
			catch (JSimInvalidProcessStateException e)
			{
				e.printStackTrace();
				e.printComment(System.err);
			} // catch
		} // synchronized(myOwnLock)
	} // finish

/* Must NOT be synchronized ! */

	/**
	 *	The process's life. Should be always overwritten in order to execute a meaningful action sequence.
	 */
	protected void life()
	{
		message(getName() + ": You should overwrite the `life()' method, sir. Read the documentation, please.");
	} // life

 /* Should be PRIVATE, but can't be */
 /* Please don't call directly this method ! */

	/**
	 *	This method should never be overwritten or called directly. It call getReady(), life() and finish() and
	 *	handles some extraordinary situations.
	 *	You should never use or modify this method (it is final).
	 */
	public final void run()
	{
		try
		{
			getReady();
			if (!shouldTerminate)
				life();
		} // try
		catch (JSimProcessDeath death)
		{
			shouldTerminate = true; // This should have been set already
			myParent.talk(getName() + " RUN: Process interrupted while sleeping.");
		} // catch
		// Catching all runtime exceptions here prevents J-Sim from hanhing up. We must run the finish()
		// method to delete the process from the system as if it finished his life normally.
		// Therefore, `shouldTerminate' must be set to false in order to switch to the main thread.
		catch (RuntimeException re)
		{
			shouldTerminate = false;
			myParent.talk("J-Sim WARNING: You have a bug in your source code!");
			myParent.talk("J-Sim WARNING: Your process `" + getName() + "' has just crashed down.");
			myParent.talk("J-Sim WARNING: The process is going to be removed from the simulation.");
			myParent.talk("Here is a description of the error:");
			myParent.talk("--ERROR-BEGINNING--");
			myParent.talk(re.toString());
			re.printStackTrace(myParent.getTalkStream());
			myParent.talk("--ERROR-END--------");
		} // catch
		finish();
	} // run

	/**
	 *	Starts the process. You should never use this method. All processes are started automatically
	 *	when the step() method of the simulation is invoked.
	 */
	public void start()
	{
		super.start();
		haveIBeenStarted = true;
	} // start

	/**
	 *	Returns true if the process has been started already. All processes are started automatically
	 *	when the step() method of the simulation is invoked. You should never use this method.
	 *
	 *	@return True if the process has been started already, false otherwise.
	 */
	protected synchronized final boolean hasBeenStarted()
	{
		return haveIBeenStarted;
	} // hasBeenStarted

	/**
	 *	Returns the process's private lock that the simulation uses for switching between itself and the process.
	 *	The lock never changes. You should never use this method.
	 *
	 *	@return	The process's private lock.
	 */
	protected Object getPrivateLock()
	{
		return myOwnLock;
	} // getPrivateLock

//--------------------------------------------------------------------------------------------------

	/**
	 *	Returns the process's number. The number is unique within the whole simulation.
	 *
	 *	@return	The process's number.
	 */
	public final long getNumber()
	{
		return myNumber;
	} // getProcessNum

	/**
	 *	Returns the process's state. Possible values are  new, passive, scheduled, active, and terminated.
	 *
	 *	@return	The process's state.
	 */
	public synchronized final int getState()
	{
		return myState;
	} // getState

	/**
	 *	Returns the time for which is the process scheduled. If the process is not scheduled, JSimCalendar.NEVER
	 *	is returned.
	 *
	 *	@return	Time for which is the process scheduled or JSimCalendar.NEVER.
	 */
	public final double getScheduleTime()
	{
		return scheduleTime;
	} // getProcessNum

	/**
	 *	Switches the process to a new state. You should never use this method since this is an internal method.
	 *
	 *	@param newState	A new state that the process has to be switched to.
	 *
	 *	@exception JSimInvalidProcessStateException	This exception is thrown out if the process is not
	 *							allowed to change its state from the current one to the
	 *							new one.
	 */
	private synchronized final void setProcessState(int newState) throws JSimInvalidProcessStateException
	{
		String errorLocation = "JSimProcess.setProcessState(): newState";
		boolean canSwitch = false;

		switch (myState)
		{
			case STATE_NEW:
				// NEW -> PASSIVE:    getReady() without activation before being started
				// NEW -> SCHEDULED:  activate() before being started (between its creation and next step())
				// NEW -> TERMINATED: simulation shutdown without the step() method being called after this process creation
				if ((newState == STATE_PASSIVE) || (newState == STATE_SCHEDULED) || (newState == STATE_TERMINATED))
					canSwitch = true;
				break; // STATE_NEW

			case STATE_PASSIVE:
				// PASSIVE -> SCHEDULED:  activate() called
				// PASSIVE -> TERMINATED: simulation shutdown when sleeping in passivate() or getReady() -- terminated via JSimProcessDeath
				if ((newState == STATE_SCHEDULED) || (newState == STATE_TERMINATED))
					canSwitch = true;
				break; // STATE_PASSIVE

			case STATE_SCHEDULED:
				// SCHEDULED -> PASSIVE:    cancel() invoked
				// SCHEDULED -> ACTIVE:     picked up from the calendar and restarted from getReady(), hold(), passivate(), or reactivate()
				// SCHEDULED -> TERMINATED: simulation shutdown when sleeping in hold() -- terminated via JSimProcessDeath
				if ((newState == STATE_PASSIVE) || (newState == STATE_ACTIVE) || (newState == STATE_TERMINATED))
					canSwitch = true;
				break; // STATE_SCHEDULED

			case STATE_ACTIVE:
				// ACTIVE -> PASSIVE:    passivate() invoked
				// ACTIVE -> SCHEDULED:  hold() invoked
				// ACTIVE -> TERMINATED: natural death -- switched in finish()
				if ((newState == STATE_PASSIVE) || (newState == STATE_SCHEDULED) || (newState == STATE_TERMINATED))
					canSwitch = true;
				break; // STATE_ACTIVE

			// Do not do anything in STATE_TERMINATED
			case STATE_TERMINATED:
				break;

			// Do not do anything in an unknown state
			default:
				break;
		} // switch

		if (canSwitch == true)
			myState = newState;
		else
			throw new JSimInvalidProcessStateException(errorLocation, myState, newState);
	} // setProcessState

	/**
	 *	Returns true if the process is idle, false otherwise. A process is idle if it has zero scheduled
	 *	events in the calendar and is not currently running, i.e. it can be activated.
	 *
	 *	@return	True if the process is idle, false otherwise.
	 */
	public boolean isIdle()
	{
		int state;

		state = getState();
		if ((state == STATE_NEW) || (state == STATE_PASSIVE))
			return true;
		else
			return false;
	} // isIdle

	/**
	 *	 Returns the process's state as string.
	 *
	 *	@return	The process's state as string.
	 */
	public String getStateAsString()
	{
		return getStateAsString(getState());
	} // getStateAsString

	/**
	 *	Returns a corresponding string representation for any given process state.
	 *
	 *	@param state	The state to be converted to a string.
	 *
	 *	@return	The given state's description.
	 */
	public static String getStateAsString(int state)
	{
		String s;

		switch (state)
		{
			case STATE_NEW:
				s = "New";
				break;
			case STATE_PASSIVE:
				s = "Passive";
				break;
			case STATE_SCHEDULED:
				s = "Scheduled";
				break;
			case STATE_ACTIVE:
				s = "Active";
				break;
			case STATE_TERMINATED:
				s = "Terminated";
				break;
			default:
				s = "Unknown";
				break;
		} // switch

		return s;
	} // getStateAsString

//--------------------------------------------------------------------------------------------------

	/**
	 *	Passivates the process. Returns the control back to the parent simulation and does not add any new
	 *	event to the calendar. A process cannot passivate another process, only itself.
	 *	Will be called from MyProcess.life() --&gt; protected because MyProcess will extend JSimProcess.
	 *	Still `a' can call `b.passivate()' but it is taken care about.
	 *
	 *	@exception JSimSecurityException	This exception is thrown out when the process to be passivated
	 *						is not currently active or if another process calls this
	 *						process's passivate() method. (Note: The two conditions should
	 *						be equivalent.)
	 *	@exception JSimProcessDeath		This internal J-Sim exception is thrown out when the simulation
	 *						is shutting down. You should NEVER catch it!
	 */
	protected final void passivate() throws JSimSecurityException
	{
		// It is not allowed to passivate somebody else.
		if (getState() != STATE_ACTIVE)
			throw new JSimSecurityException("JSimProcess.passivate(): Only an active process can be passivated.");
		if (myParent.getRunningProcess() != this)
			throw new JSimSecurityException("JSimProcess.passivate(): You may not do this!");

		synchronized(myOwnLock)
		{
			try // outer try
			{
				if ((shouldTerminate) && (requestsToTerminate >= INT_REQUESTS_TO_IGNORE))
				{
					myParent.talk("J-Sim WARNING: You probably suppress J-Sim's mechanisms of process handling.");
					myParent.talk("J-Sim WARNING: This is very dangerous and may lead to deadlock.");
					myParent.talk("J-Sim WARNING: Please check up your source code unless you really know");
					myParent.talk("J-Sim WARNING: what you are doing.");
				} // if

				// Wake up main() which is now waiting in step().
				myParent.switchToNobody();
				myOwnLock.notify();

				// Going to sleep now.
				setProcessState(STATE_PASSIVE);
				while (myParent.getRunningProcess() != this)
				{
					try // inner try
					{
						if (shouldTerminate)
							requestsToTerminate++;
						myOwnLock.wait();
					} // inner try
					catch (InterruptedException e1)
					{
						shouldTerminate = true;
						requestsToTerminate++;
						// Why `break' is used?
						// We must get out of the cycle and throw out the exception.
						// We can't simply test (!shouldTerminate) in the cycle header
						// because we would not ever sleep again.
						// This means the current process would run without any hold()s
						// or passivate()s to infinity...
						break;
					} // catch
				} // while not me

				// If we have to die this exception is propagated till run().
				if (shouldTerminate)
				{
					myParent.talk(getName() + " PASSIVATE: Killed.");
					throw new JSimProcessDeath();
				} // if
				else
				{
					setProcessState(STATE_ACTIVE);
					scheduleTime = JSimCalendar.NEVER;
				} // else
			} // outer try
			catch (JSimInvalidProcessStateException e2)
			{
				e2.printStackTrace();
				e2.printComment(System.err);
			} // catch
			// We are now normally returning to our life().
		} // synchronized
	} // passivate

	/**
	 *	Passivates the process but adds a new entry to the calendar, specifying the time when the process will
	 *	be activated again. A process cannot call another process's hold(), only its own.
	 *	Will be called from MyProcess.life() --&gt; protected because MyProcess will extend JSimProcess.
	 *	Still `a' can call `b.hold()' but it is taken care about.
	 *
	 *	@param deltaT	The time delta specifying the difference between the current time and the time when
	 *			the process will be activated.
	 *
	 *	@exception JSimSecurityException	This exception is thrown out when the process to be hold is not
	 *						currently active or if another process calls this process's
	 *						hold() method. (Note: The two conditions should be equivalent.)
	 *	@exception JSimProcessDeath		This internal J-Sim exception is thrown out when the simulation
	 *						is shutting down. You should NEVER catch it!
	 */
	protected final void hold(double deltaT) throws JSimSecurityException
	{
		// It is not allowed to hold somebody else.
		if (getState() != STATE_ACTIVE)
			throw new JSimSecurityException("JSimProcess.hold(): Only an active process can be hold.");
		if (myParent.getRunningProcess() != this)
			throw new JSimSecurityException("JSimProcess.hold(): You may not do this!");

		synchronized(myOwnLock)
		{
			try // outer try
			{
				scheduleTime = myParent.getCurrentTime() + Math.abs(deltaT);
				myParent.addEntryToCalendar(scheduleTime, this);
				if ((shouldTerminate) && (requestsToTerminate >= INT_REQUESTS_TO_IGNORE))
				{
					myParent.talk("J-Sim WARNING: You probably suppress J-Sim's mechanisms of process handling.");
					myParent.talk("J-Sim WARNING: This is very dangerous and may lead to deadlock.");
					myParent.talk("J-Sim WARNING: Please check up your source code unless you really know");
					myParent.talk("J-Sim WARNING: what you are doing.");
				} // if

				// Wake up main() which is waiting now in step().
				myParent.switchToNobody();
				myOwnLock.notify();

				// Going to sleep.
				setProcessState(STATE_SCHEDULED);
				while (myParent.getRunningProcess() != this)
				{
					try // inner try
					{
						if (shouldTerminate)
							requestsToTerminate++;
						myOwnLock.wait();
					} // inner try
					catch (InterruptedException e1)
					{
						shouldTerminate = true;
						requestsToTerminate++;
						// Why `break' is used?
						// We must get out of the cycle and throw out the exception.
						// We can't simply test (!shouldTerminate) in the cycle header
						// because we would not ever sleep again.
						// This means the current process would run without any hold()s
						// or passivate()s to infinity...
						break;
					} // catch
				} // while

				// If we have to die this exception is propagated till run().
				if (shouldTerminate)
				{
					myParent.talk(getName() + " HOLD: Killed.");
					throw new JSimProcessDeath();
				} // if
				else
				{
					setProcessState(STATE_ACTIVE);
					scheduleTime = JSimCalendar.NEVER;
				} // else
			} // outer try
			catch (JSimInvalidParametersException e2) // from addEntryToCalendar
			{
				e2.printStackTrace();
				e2.printComment(System.err);
			} // catch
			catch (JSimInvalidProcessStateException e3) // from setProcessState
			{
				e3.printStackTrace();
				e3.printComment(System.err);
			} // catch
			// We are now returning to our life().
		} // synchronized
	} // hold

	/**
	 *	Activates a process at the given time. Adds a new entry to the calendar, specifying the time when
	 *	the process will be activated. A process can activate itself or any other process.
	 *	Will be called from MySim.main() or MyProcess.life() --&gt; public because MySim will not be
	 *	in the same package.
	 *
	 *	@param when	Absolute simulation time when the process should be activated.
	 *
	 *	@exception JSimSecurityException		This exception is thrown out when the process to be
	 *							activated is not currently passive or new.
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the time of activation
	 *							is less than current simulation time.
	 */
	public final void activate(double when) throws JSimSecurityException, JSimInvalidParametersException
	{
		if (! ((getState() == STATE_NEW) || (getState() == STATE_PASSIVE)) )
			throw new JSimSecurityException("JSimProcess.activate(): Only a passive or new process can be activated.");

		synchronized(myOwnLock)
		{
			if (when >= myParent.getCurrentTime())
			{
				try
				{
					myParent.addEntryToCalendar(when, this);
					scheduleTime = when;
					setProcessState(STATE_SCHEDULED);
				} // try
				catch (JSimInvalidParametersException e1) // from addCalendarEntry
				{
					e1.printStackTrace();
					e1.printComment(System.err);
				} // catch
				catch (JSimInvalidProcessStateException e2) // from setProcessState
				{
					e2.printStackTrace();
					e2.printComment(System.err);
				} // catch
			} // if when >= current time
			else
				throw new JSimInvalidParametersException("JSimProcess.activate(): when (less than current simulation time)");
		} // synchronized
	} // activate

	/**
	 *	Deletes all process's scheduled events from the calendar. The process will not be run anymore unless
	 *	activated again by another process. A process can cancel another process but not itself.
	 *	Will be called from MyProcess.life() --&gt; public because it need not necessarily be in the same package
	 *	as the called process.
	 *
	 *	@exception JSimSecurityException	This exception is thrown out when the process to be cancelled is
	 *						not currently scheduled or if the process calls its own cancel()
	 *						method. (Note: The latter condition is a subset of the former one
	 *						because the currently running process can never be scheduled.)
	 */
	public final void cancel() throws JSimSecurityException
	{
		// A process can only cancel a another process that is scheduled.
		if (getState() != STATE_SCHEDULED)
			throw new JSimSecurityException("JSimProcess.cancel(): Only a scheduled process can be cancelled.");
		if (myParent.getRunningProcess() == this)
			throw new JSimSecurityException("JSimProcess.cancel(): You may not cancel yourself!");

		synchronized(myOwnLock)
		{
			try
			{
				myParent.deleteEntriesInCalendar(this, true);
				setProcessState(STATE_PASSIVE);
				scheduleTime = JSimCalendar.NEVER;
			} // try
			catch (JSimInvalidParametersException e1) // from deleteEntriesInCalendar
			{
				e1.printStackTrace();
				e1.printComment(System.err);
			} // catch
			catch (JSimInvalidProcessStateException e2) // from setProcessState
			{
				e2.printStackTrace();
				e2.printComment(System.err);
			} // catch
		} // synchronized
	} // cancel

	/**
	 *	Reactivates the process. The process is temporarily suspended as with passivate() but an event is
	 *	added to the calendar for the same simulation time. This means that the process will run again
	 *	in a future step (not necessarily the next one) and the value of the simulation time will be the same
	 *	as now. The purpose of this method is to let other processes run if they have been scheduled for
	 *	the same simulation time. Since later scheduled processes are run later (if the simulation times are equal)
	 *	this process will run after all processes already scheduled for the current simulation time.
	 *
	 *	@exception JSimSecurityException	This exception is thrown out when the process to be reactivated
	 *						is not currently active or if another process calls this process's
	 *						hold() method. (Note: The two conditions should be equivalent.)
	 *	@exception JSimProcessDeath		This internal J-Sim exception is thrown out when the simulation
	 *						is shutting down. You should NEVER catch it!
	 */
	protected final void reactivate() throws JSimSecurityException
	{
		// It is not allowed to reactivate somebody else.
		if (getState() != STATE_ACTIVE)
			throw new JSimSecurityException("JSimProcess.reactivate(): Only an active process can be hold.");
		if (myParent.getRunningProcess() != this)
			throw new JSimSecurityException("JSimProcess.reactivate(): You may not do this!");

		hold(0.0);
	} // reactivate

//--------------------------------------------------------------------------------------------------

	/**
	 *	Prints out a text info message, either to the standard output or to the simulation window.
	 *
	 *	@param s	The message to be printed out.
	 */
	public void message(String s)
	{
		myParent.printString(s, JSimSimulation.PRN_MESSAGE, true);
	} // message

	/**
	 *	Prints a text info message, either to the standard output or to the simulation window, but
	 *	does not teminate the line.
	 *
	 *	@param s	The message to be printed out.
	 */
	public void messageNoNL(String s)
	{
		myParent.printString(s, JSimSimulation.PRN_MESSAGE, false);
	} // message

	/**
	 *	Prints out a text error message, either to the error output or to the simulation window.
	 *
	 *	@param s	The error message to be printed out.
	 */
	public void error(String s)
	{
		myParent.printString(s, JSimSimulation.PRN_ERROR, true);
	} // error

//--------------------------------------------------------------------------------------------------

	/**
	 *	Returns a string containing basic information about the process. The string can be displayed
	 *	in a JSimMainWindowList component.
	 *
	 *	@return	A string containing basic information about the process.
	 */
	public String getObjectListItemDescription()
	{
		return toString();
	} // getObjectListItemDescription

	/**
	 *	Returns a collection of process's characteristics. Every characteristics contains a name
	 *	and a value. The collection can be displayed in a JSimDetailedInfoWindow table.
	 *
	 *	@return	A collection of JSimPairs.
	 */
	public Collection getDetailedInformationArray()
	{
		Collection c = new ArrayList(5);
		c.add(new JSimPair("Number:", Long.toString(getNumber())));
		c.add(new JSimPair("Name:", getName()));
		c.add(new JSimPair("State:", getStateAsString()));
		if (myState == STATE_SCHEDULED)
			c.add(new JSimPair("Scheduled for:", Double.toString(scheduleTime)));
		return c;
	} // getDetailedInformationArray

	/**
	 *	Creates a detailed info window that shows information about this process.
	 *	Returns a reference to the created window.
	 *
	 *	@return	Reference to the created info window.
	 */
	public JDialog createDetailedInfoWindow(JSimMainWindow parentWindow)
	{
		JSimDetailedInfoWindow dWindow = new JSimDetailedInfoWindow(parentWindow, this);
		return dWindow;
	} // createDetailedInfoWindow

//--------------------------------------------------------------------------------------------------

	/**
	 *	Returns a string representation of the process. Provided information: number, name, state,
	 *	and schedule time.
	 *
	 *	@return	A string representation of the process.
	 */
	public String toString()
	{
		String s = "No.: " + getNumber() + " Name: " + getName() + " State: " + getStateAsString();
		if (getState() == STATE_SCHEDULED)
			return (s + " for " + getScheduleTime());
		else
			return s;
	} // toString

	/**
	 *	Compares this process with another one. Returns a negative integer, zero, or a positive integer
	 *	as this process is less than, equal to, or greater than the specified object. It is assumed that
	 *	the argument is also a JSimProcess.
	 *	This class has a natural ordering that is fully consistent with equals(). If equals() returns true
	 *	for p1 and p2, then compareTo() will return 0 for the same p1 and p2, and vice versa.
	 *
	 *	@return	Zero if the numbers of both processes are equal, a negative number if the number of this
	 *		process is less than the other process's number, and a positive number if the number of this
	 *		process is greater than the other process's number.
	 *
	 *	@exception ClassCastException	This exception is thrown out when the specified object cannot be
	 *					typecasted to JSimProcess.
	 */
	public int compareTo(Object o)
	{
		JSimProcess p = (JSimProcess) o;

		if (this.myParent.getNumber() == p.myParent.getNumber())
			if (this.myNumber == p.myNumber)
				return 0;
			else
				if (this.myNumber < p.myNumber)
					return -1;
				else
					return +1;
		else
			if (this.myParent.getNumber() < p.myParent.getNumber())
				return -1;
			else
				return +1;
	} // compareTo

	/**
	 *	Indicates whether some other object is equal to this one. This implementation compares process numbers
	 *	and their simulations' numbers which is actually equal to simple reference comparison because
	 *	process numbers are unique for a given simulation and simulation numbers are unique for a given JVM
	 *	instance.
	 *	Unique process numbers are assured by the constructor and the JSimSimulation.getFreeProcessNumber()
	 *	method. Unique simulation numbers are assured by the JSimSimulation constructor.
	 *
	 *	@param o	The reference object with which to compare.
	 *
	 *	@return	True if this object is the same as the obj argument, false otherwise.
	 */
	public boolean equals(Object o)
	{
		if (o == this)
			return true;

		if ((o instanceof JSimProcess) == false)
			return false;

		JSimProcess p = (JSimProcess) o;

		if ((this.myNumber == p.myNumber) && (this.myParent.getNumber() == p.myParent.getNumber()))
			return true;
		else
			return false;
	} // equals

	/**
	 *	Returns a hash code value for the object. The hash code is computed from the process's number and its
	 *	simulation's number using the algorithm described in [UJJ3/166]. This implementation of hash code
	 *	computation is fully consistent with equals().
	 *
	 *	@return	A hash code for this process.
	 */
	public int hashCode()
	{
		int temp = 17; // Magic number 17
		int myNumberAsInt = (int) (myNumber ^ (myNumber >>> 32));
		int simulationNumber = myParent.getNumber();

		temp = 37 * temp + myNumberAsInt; // Another magic number 37
		temp = 37 * temp + simulationNumber;

		return temp;
	} // hashCode

} // class JSimProcess

