/*
 *	Copyright (c) 2000 - 2004, University of West Bohemia,
 *	                           Faculty of Applied Sciences,
 *	                           Department of Computer Science and Engineering,
 *	                           DSS Group.
 *	                           Pilsen, Czech Republic.
 *
 *	Please read the licence agreement carefully before using any part of this software.
 */
package cz.zcu.fav.kiv.jsim.unsecure;

import cz.zcu.fav.kiv.jsim.*;

import java.util.*;

/**
 *	UnsecureSimulation is an <em>experimental</em> class that you probably will not use at all. It is intended
 *	to be used in distributed simulation where simulation time rollbacks can occur. This class allows you to change
 *	the simulation time arbitrarily (even backwards) and to access calendar events for reading before they are
 *	interpreted. This can be used for future forecast, e.g. for computation of the lookahead time.
 *	You had better not use this class at all.
 *
 *	@author		<a href="http://home.zcu.cz/~jkacer">Jarda KAER</a>
 *
 *	@version	J-Sim version 0.3.0
 *
 *	@since		J-Sim version 0.3.0
 */
public class UnsecureSimulation extends JSimSimulation
{
	/**
	 *	A flag saying that the simulation will have to roll back.
	 */
	private boolean undoComing;

	/**
	 *	Creates a new simulation with no processes, no queues and no graphic window. The simulation will
	 *	run in text (console) mode and no graphic output will be allowed.
	 *
	 *	@param name	Name of the simulation.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the specified name is
	 *							is null.
	 */
	public UnsecureSimulation(String name) throws JSimInvalidParametersException
	{
		this(name, JSimSimulation.MODE_TXT, 0.0);
	} // constructor

	/**
	 *	Creates a new simulation with no processes, no queues and no graphic window. According to the
	 *	simulation mode specified as a parameter, it will / will not be allowed to use graphic output later.
	 *	If GUI batch or interactive mode is used, the output will be updated after every step.
	 *
	 *	@param name	Name of the simulation.
	 *	@param mode	Simulation mode. Possible values are MODE_TXT, MODE_GUI_BATCH,
	 *			and MODE_GUI_INTERACTIVE.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the specified name is
	 *							is null or an invalid simulation mode is given to the
	 							constructor.
	 */
	public UnsecureSimulation(String name, int mode) throws JSimInvalidParametersException
	{
		this(name, mode, 0.0);
	} // constructor

	/**
	 *	Creates a new simulation with no processes, no queues and no graphic window. According to the
	 *	simulation mode specified as a parameter, it will / will not be allowed to use graphic output later.
	 *
	 *	@param name		Name of the simulation.
	 *	@param mode		Simulation mode. Possible values are MODE_TXT, MODE_GUI_BATCH,
	 *				and MODE_GUI_INTERACTIVE.
	 *	@param guiOutputDelta	A simulation time interval after which the GUI output will be updated. If set to
	 *				zero, the output will be updated after every simulation step. This may be
	 *				time consuming and inappropriate for some situations.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the specified name is
	 *							is null, an invalid simulation mode is given to the
	 							constructor, or  if the update interval is less than zero.
	 */
	public UnsecureSimulation(String name, int mode, double guiUpdateDelta) throws JSimInvalidParametersException
	{
		super(name, mode, guiUpdateDelta);
		undoComing = false;
	} // constructor

	/**
	 *	Reports the status of the undo-coming (future rollback) flag.
	 *
	 *	@return	True if the simulation will have to roll back, false otherwise.
	 */
	public boolean isUndoComing()
	{
		return undoComing;
	} // isUndoComing

	/**
	 *	Changes the status of the undo-coming (future rollback) flag.
	 *
	 *	@param newUndoComing	The new value of the undo-coming flag.
	 */
	protected void setUndoComing(boolean newUndoComing)
	{
		undoComing = newUndoComing;
	} // setUndoComing

	/**
	 *	Advances the simulation time by a positive delta.
	 *
	 *	@param positiveDelta	A simulation time difference between the current and desired simulation time.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the delta is negative.
	 */
	protected void advanceTime(double positiveDelta) throws JSimInvalidParametersException
	{
		if (positiveDelta < 0.0)
			throw new JSimInvalidParametersException("UnsecureSimulation.advanceTime(): positiveDelta");
		time += positiveDelta;
	} // advanceTime

	/**
	 *	Rolls back the simulation time by a negative delta.
	 *
	 *	@param negativeDelta	A simulation time difference between the current and desired simulation time.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the delta is positive.
	 */
	protected void rollbackTime(double negativeDelta) throws JSimInvalidParametersException
	{
		if (negativeDelta > 0.0)
			throw new JSimInvalidParametersException("UnsecureSimulation.rollbackTime(): negativeDelta");
		time += negativeDelta;
	} // rollbackTime

	/**
	 *	Sets the value of the simulation time to a new value.
	 *
	 *	@param absoluteTime	The new value of simulation time.
	 */
	protected void setTime(double absoluteTime)
	{
		time = absoluteTime;
	} // setTime

	/**
	 *	Adds a new event to the calendar. The event must be valid, i.e. its process must not be null
	 *	and its simulation time must not be negative. You are fully responsible for correct creation
	 *	of the event. Adding an incorrect event to the calendar may lead to an unpredicted behavior.
	 *	Pay attention especially to the following:
	 *	<ol>
	 *		<li>
	 *		The simulation time of the event should not be less than the current simulation time.
	 *		If it is, your simulation will jump to the past.
	 *		</li>
	 *		<li>
	 *		The event's process should be a valid J-Sim process present in the same simulation as the calendar.
	 *		</li>
	 *	</ol>
	 *
	 *	@param jsce	The event to be added to the calendar.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the event is null or it is
	 *							rejected by the calendar.
	 */
	protected void addWholeEventToCalendar(JSimCalendarEvent jsce) throws JSimInvalidParametersException
	{
		if (jsce == null)
			throw new JSimInvalidParametersException("UnsecureSimulation.addWholeEventToCalendar(): jsce");

		calendar.addWholeEvent(jsce);
	} // addWholeEntryToCalendar

	/**
	 *	Returns copies of first N events from the calendar. You can modify them at your discretion since
	 *	they are just copies of the original events. If there are less events in the calendar than you
	 *	specified (N), some positions in the resulting array will be null. A valid (not-null) array will
	 *	always be returned, even if there are no events in the calendar.
	 *
	 *	@param n	The number of events to be returned.
	 *
	 *	@return	An array of N first events from the calendar, some of which can be null.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if N is less than 1.
	 */
	protected JSimCalendarEvent[] getFirstNEventsFromCalendar(int n) throws JSimInvalidParametersException
	{
		JSimCalendarEvent[] array;
		int i = 0;
		ListIterator it;

		if (n < 1)
			throw new JSimInvalidParametersException("UnsecureSimulation.getFirstNEventsFromCalendar(): n");

		array = new JSimCalendarEvent[n];
		for (i = 0; i < array.length; i++)
			array[i] = null;

		it = calendar.getIterator();
		while (it.hasNext() && (i < n))
		{
			array[i] = (JSimCalendarEvent) it.next();
			i++;
		} // while

		return array;
	} // getFirstNEventsFromCalendar

	/**
	 *	Returns a list iterator over copies of all events of the calendar. You may modify them at your
	 *	discretion without affecting the calendar itself.
	 *
	 *	@return	A list iterator over copies of all events of the calendar.
	 */
	protected ListIterator getAllEventsFromCalendar()
	{
		return calendar.getIterator();
	} // getAllEventsFromCalendar

} // class UnsecureSimulation

